from .result import ODEResult
from .solver import (
    classical_scheme, 
    Euler_scheme, 
    trigonometric_scheme
)

__all__ = [
    "classical_scheme", 
    "Euler_scheme", 
    "trigonometric_scheme", 
    "ODEResult"
]


import os
import pandas as pd
import numpy as np
import matplotlib.pyplot as plt
from typing import Optional, Callable, Tuple

class ODEResult:
    def __init__(
        self, 
        time_grid: np.ndarray, 
        numerical_sol: np.ndarray, 
        scheme: str
    ):
        self.time_grid = time_grid
        self.numerical_sol = numerical_sol
        self.scheme = scheme
    
    def values(
        self
    ) -> Tuple[np.ndarray, np.ndarray]:
        """
        Returns the time grid and the numerical solution as a tuple.
        
        Returns
        -------
        Tuple[np.ndarray, np.ndarray]
            (time_grid, numerical_sol)
        """
        return self.time_grid, self.numerical_sol
    
    def visualization(
        self, 
        figure_size: tuple = (5.5, 2.5), 
        exact_sol: Optional[Callable[[float], float]] = None,                       
        save_path: Optional[str] = None
    ):
        plt.figure(figsize=figure_size)
        
        if exact_sol is not None:
            exact_values = np.array([exact_sol(t) for t in self.time_grid])
            plt.plot(self.time_grid, exact_values, color='black', label='Exact solution')

        number_of_circles = max(1, len(self.time_grid) // 30)

        capitalized_name = self.scheme[0].upper() + self.scheme[1:]
        
        plt.plot(self.time_grid, self.numerical_sol, linestyle='--', marker='o', color='red', markersize=5, markevery=number_of_circles, markerfacecolor='none', markeredgewidth=1.0, label=capitalized_name)

        plt.xlabel(r"Time", fontsize=10)
        plt.ylabel(r"Solution", fontsize=10)
        plt.grid(True)
        plt.legend(loc='center left', bbox_to_anchor=(1.02, 0.5), borderaxespad=0., fontsize=10)

        if save_path:
            if not os.path.exists('figures'):
                os.makedirs('figures')
            
            save_path = save_path.replace(" ", "-")
            full_path = os.path.join("figures", save_path)

            if not save_path.endswith(".png"):
                save_path += ".png"
                
            plt.savefig(full_path, dpi=1000, bbox_inches='tight')

            print(f"Figure saved: {full_path}")
        
        plt.show()

        return self

    def table(self,
        exact_sol: Optional[Callable[[float], float]] = None,   
        save_path: Optional[str] = None
    ):
        
        result = pd.DataFrame(self.numerical_sol, index=self.time_grid, columns=["Numerical solution"])
        result.index.name = "Time"

        if exact_sol:
            result["Exact solution"] = [exact_sol(t) for t in self.time_grid]
            result["Error"] = abs(result["Numerical solution"] - result["Exact solution"])

        if save_path:
            if not os.path.exists('tables'):
                os.makedirs('tables')

            save_path = save_path.replace(" ", "-")
            full_path = os.path.join("tables", save_path)
            
            if full_path.endswith(".txt"):
                with open(full_path, "w") as f:
                    f.write(result.to_string())
            else:
                if not full_path.endswith(".csv"):
                    full_path += ".csv"
                
                result.to_csv(full_path) 
            
            print(f"Table saved: {full_path}")

        return result

def save_table_to_txt(
    df: pd.DataFrame,
    filename: str,
    error_precision: int = 2,
    ratio_precision: int = 2
) -> None:
    """
    Saves a DataFrame of convergence results to a text file in LaTeX table format.

    Parameters
    ----------
    df : pd.DataFrame
        A DataFrame with three columns: 'n', 'error', and 'ratio'.
    filename : str
        The name of the text file to save the results to.
    error_precision : int, optional
        The number of decimal places for the 'error' column. 
        Default: ``2``.
    ratio_precision : int, optional
        The number of decimal places for the 'ratio' column. 
        Default: ``2``.

    Returns
    -------
    None
        This function does not return any value; it writes directly to a file.
    """
    if not os.path.exists('tables'):
        os.makedirs('tables')

    filename = "tables/" + filename

    with open(filename, "w") as f:
        for n, error, ratio in df.itertuples(index=False, name=None):
            error_str = f"{error:.{error_precision}e}"
            ratio_str = f"{ratio:.{ratio_precision}f}" if pd.notna(ratio) else "--"
            f.write(f"{n:<8}& {error_str} & {ratio_str} \\\\\n")



import numpy as np
from tqdm import tqdm
import time
from typing import Callable, List, Tuple, Optional
from .result import OptimizationResult
from .step_size import StepSize
from ..calculation import derivative

SUPPORTED_METHODS = ['specular gradient method', '', 'implicit', 'stochastic', 'hybrid']

def gradient_method(
    method: str,
    objective_func: Callable[[int | float | list | np.ndarray], int | float | np.floating],
    initial_point: int | float | list | np.ndarray, 
    step_size: StepSize,  
    max_iteration: int = 1000, 
    tol: float = 1e-6, 
    h: Optional[float] = 1e-6, 
    first_iteration: Optional[int] = 2,   
    name: Optional[str] = "no name",
    record_history: bool = True,
    record_time: bool = True
) -> OptimizationResult:

    x = np.array(initial_point, dtype=float).copy()
    f = objective_func
    n = x.ndim
    k = 0
    
    history = {}
    x_history = []
    f_history = []

    if record_time is True:
        start_time = time.time()

    # the n-dimensional case
    if n > 0:
        if method == '' or method == 'specular gradient method':
            pass # TODO

        else:
            raise ValueError(f"Unknown method '{method}'. Supported methods: {SUPPORTED_METHODS}")

    # the one-dimensional case
    else:
        x = x.item()

        if method == '' or method == 'specular gradient method':
            if h is None or h <= 0:
                raise ValueError("Numerical differentiation requires a positive step size 'h'.")
            
            print(f"[Specular gradient method]    {name}")

            for _ in tqdm(range(1, max_iteration + 1)):
                if record_history is True:
                    x_history.append(x)
                    f_history.append(f(x))

                specular_derivative = derivative(f=f, x=x, h=h) # type: ignore

                if abs(specular_derivative) < tol:
                    break

                k += 1
                x -= step_size(k)*(specular_derivative / abs(specular_derivative))

        elif method == 'implicit':
            if h is None or h <= 0:
                raise ValueError("Numerical differentiation requires a positive step size 'h'.")
            
            print(f"[Implicit specular gradient method]    {name}")

            for _ in tqdm(range(1, max_iteration + 1)):
                if record_history is True:
                    x_history.append(x)
                    f_history.append(f(x))

                sum_of_one_sided_derivatives = (f(x + h) - f(x - h)) / h

                if abs(sum_of_one_sided_derivatives) < tol:
                    break
                
                k += 1
                x -= step_size(k)*(sum_of_one_sided_derivatives / abs(sum_of_one_sided_derivatives))
        else:
            raise ValueError(f"Unknown method '{method}'. Supported methods: {SUPPORTED_METHODS}")

    print("\n")
    
    if record_history is True:
        history["variables"] = x_history
        history["values"] = f_history

    if record_time is True:
        history["time"] = time.time() - start_time  # type: ignore

    history["method"] = method
    history["name"] = name

    return OptimizationResult(solution=x, objective_func_val=f(x), iteration=k, history=history) # type: ignore


# def _specular_gradient_method_scalar(
#     method: str,
#     objective_func: Callable[[int | float | list | np.ndarray], int | float | np.floating],
#     initial_point: int | float | list | np.ndarray, 
#     step_size: StepSize,  
#     max_iteration: int = 1000, 
#     tol: float = 1e-6, 
#     h: Optional[float] = 1e-6, 
#     first_iteration: Optional[int] = 2,   
#     name: Optional[str] = "no name",
#     record_history: bool = True,
#     record_time: bool = True):

# def _specular_gradient_method_vector(
#     method: str,
#     objective_func: Callable[[int | float | list | np.ndarray], int | float | np.floating],
#     initial_point: int | float | list | np.ndarray, 
#     step_size: StepSize,  
#     max_iteration: int = 1000, 
#     tol: float = 1e-6, 
#     h: Optional[float] = 1e-6, 
#     first_iteration: Optional[int] = 2,   
#     name: Optional[str] = "no name",
#     record_history: bool = True,
#     record_time: bool = True):